@@ -45,6 +45,7 @@
 #include <osmo-bts/signal.h>
 #include <osmo-bts/l1sap.h>
 #include <osmo-bts/oml.h>
+#include <osmo-bts/abis_osmo.h>
 
 uint32_t trx_get_hlayer1(const struct gsm_bts_trx *trx);
 
@@ -62,8 +63,6 @@ static const char *sapi_string[] = {
 	[PCU_IF_SAPI_PTCCH] = 	"PTCCH",
 };
 
-static int pcu_sock_send(struct gsm_network *net, struct msgb *msg);
-
 /*
  * PCU messages
  */
@@ -100,7 +99,7 @@ static bool ts_should_be_pdch(const struct gsm_bts_trx_ts *ts)
 			return !(ts->flags & TS_F_PDCH_DEACT_PENDING);
 		else
 			return (ts->flags & TS_F_PDCH_ACT_PENDING);
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		/*
 		 * When we're busy de-/activating the PDCH, we first set
 		 * ts->dyn.pchan_want, tell the PCU about it and wait for a
@@ -217,7 +216,7 @@ static void info_ind_fill_trx(struct gsm_pcu_if_info_trx *trx_info,
 			continue;
 
 		trx_info->pdch_mask |= (1 << tn);
-		trx_info->ts[tn].tsc = gsm_ts_tsc(ts);
+		trx_info->ts[tn].tsc = ts->tsc;
 
 		if (ts->hopping.enabled)
 			info_ind_fill_fhp(&trx_info->ts[tn], ts);
@@ -266,6 +265,7 @@ int pcu_tx_info_ind(void)
 	if (pcu_direct)
 		info_ind->flags |= PCU_IF_FLAG_SYSMO;
 
+	info_ind->bsic = bts->bsic;
 	/* RAI */
 	info_ind->mcc = net->plmn.mcc;
 	info_ind->mnc = net->plmn.mnc;
@@ -501,7 +501,8 @@ int pcu_tx_data_ind(struct gsm_bts_trx_ts *ts, uint8_t sapi, uint32_t fn,
 	data_ind->ber10k = ber10k;
 	data_ind->ta_offs_qbits = bto;
 	data_ind->lqual_cb = lqual;
-	memcpy(data_ind->data, data, len);
+	if (len)
+		memcpy(data_ind->data, data, len);
 	data_ind->len = len;
 
 	return pcu_sock_send(&bts_gsmnet, msg);
@@ -558,6 +559,39 @@ int pcu_tx_time_ind(uint32_t fn)
 	return pcu_sock_send(&bts_gsmnet, msg);
 }
 
+int pcu_tx_interf_ind(const struct gsm_bts_trx *trx, uint32_t fn)
+{
+	struct gsm_pcu_if_interf_ind *interf_ind;
+	struct gsm_pcu_if *pcu_prim;
+	struct msgb *msg;
+	unsigned int tn;
+
+	msg = pcu_msgb_alloc(PCU_IF_MSG_INTERF_IND, trx->bts->nr);
+	if (!msg)
+		return -ENOMEM;
+	pcu_prim = (struct gsm_pcu_if *) msg->data;
+	interf_ind = &pcu_prim->u.interf_ind;
+
+	interf_ind->trx_nr = trx->nr;
+	interf_ind->fn = fn;
+
+	for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+		const struct gsm_bts_trx_ts *ts = &trx->ts[tn];
+		const struct gsm_lchan *lchan = &ts->lchan[0];
+
+		if (ts->mo.nm_state.operational != NM_OPSTATE_ENABLED)
+			continue;
+		if (ts->mo.nm_state.availability != NM_AVSTATE_OK)
+			continue;
+		if (ts_pchan(ts) != GSM_PCHAN_PDCH)
+			continue;
+
+		interf_ind->interf[tn] = -1 * lchan->meas.interf_meas_avg_dbm;
+	}
+
+	return pcu_sock_send(&bts_gsmnet, msg);
+}
+
 int pcu_tx_pag_req(const uint8_t *identity_lv, uint8_t chan_needed)
 {
 	struct pcu_sock_state *state = bts_gsmnet.pcu_state;
@@ -651,7 +685,7 @@ static int pcu_rx_data_req(struct gsm_bts *bts, uint8_t msg_type,
 
 	switch (data_req->sapi) {
 	case PCU_IF_SAPI_PCH:
-		paging_add_imm_ass(bts->paging_state, data_req->data, data_req->len);
+		paging_add_imm_ass(bts->paging_state, data_req->data, data_req->len, true);
 		break;
 	case PCU_IF_SAPI_AGCH:
 		msg = msgb_alloc(data_req->len, "pcu_agch");
@@ -768,8 +802,8 @@ int pcu_tx_si(const struct gsm_bts *bts, enum osmo_sysinfo_type si_type,
 
 static int pcu_tx_si_all(struct gsm_bts *bts)
 {
-	const enum osmo_sysinfo_type si_types[3] =
-	    { SYSINFO_TYPE_1, SYSINFO_TYPE_3, SYSINFO_TYPE_13 };
+	const enum osmo_sysinfo_type si_types[] =
+	    { SYSINFO_TYPE_1, SYSINFO_TYPE_2, SYSINFO_TYPE_3, SYSINFO_TYPE_13 };
 	unsigned int i;
 	int rc = 0;
 
@@ -845,7 +879,7 @@ static int pcu_rx_act_req(struct gsm_bts *bts,
 		     gsm_lchant_name(lchan->type));
 		return -EINVAL;
 	}
-	if (lchan->ts->pchan == GSM_PCHAN_TCH_F_TCH_H_PDCH &&
+	if (lchan->ts->pchan == GSM_PCHAN_OSMO_DYN &&
 	    lchan->ts->dyn.pchan_is != GSM_PCHAN_PDCH) {
 		LOGP(DPCU, LOGL_ERROR,
 		     "%s request, but lchan in dyn TS is not configured as PDCH in lower layers (is %s)\n",
@@ -861,11 +895,21 @@ static int pcu_rx_act_req(struct gsm_bts *bts,
 	return 0;
 }
 
+#define CHECK_IF_MSG_SIZE(prim_len, prim_msg) \
+	do { \
+		size_t _len = PCUIF_HDR_SIZE + sizeof(prim_msg); \
+		if (prim_len < _len) { \
+			LOGP(DPCU, LOGL_ERROR, "Received %zu bytes on PCU Socket, but primitive %s " \
+			     "size is %zu, discarding\n", prim_len, #prim_msg, _len); \
+			return -EINVAL; \
+		} \
+	} while(0);
 static int pcu_rx(struct gsm_network *net, uint8_t msg_type,
-	struct gsm_pcu_if *pcu_prim)
+	struct gsm_pcu_if *pcu_prim, size_t prim_len)
 {
 	int rc = 0;
 	struct gsm_bts *bts;
+	size_t exp_len;
 
 	if ((bts = gsm_bts_num(net, pcu_prim->bts_nr)) == NULL) {
 		LOGP(DPCU, LOGL_ERROR, "Received PCU Prim for non-existent BTS %u\n", pcu_prim->bts_nr);
@@ -874,17 +918,32 @@ static int pcu_rx(struct gsm_network *net, uint8_t msg_type,
 
 	switch (msg_type) {
 	case PCU_IF_MSG_DATA_REQ:
+		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.data_req);
 		rc = pcu_rx_data_req(bts, msg_type, &pcu_prim->u.data_req);
 		break;
 	case PCU_IF_MSG_PAG_REQ:
+		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.pag_req);
 		rc = pcu_rx_pag_req(bts, msg_type, &pcu_prim->u.pag_req);
 		break;
 	case PCU_IF_MSG_ACT_REQ:
+		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.act_req);
 		rc = pcu_rx_act_req(bts, &pcu_prim->u.act_req);
 		break;
 	case PCU_IF_MSG_TXT_IND:
+		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.txt_ind);
 		rc = pcu_rx_txt_ind(bts, &pcu_prim->u.txt_ind);
 		break;
+	case PCU_IF_MSG_CONTAINER:
+		CHECK_IF_MSG_SIZE(prim_len, pcu_prim->u.container);
+		/* ^ check if we can access container fields, v check with container data length */
+		exp_len = PCUIF_HDR_SIZE + sizeof(pcu_prim->u.container) + osmo_load16be(&pcu_prim->u.container.length);
+		if (prim_len < exp_len) {
+			LOGP(DPCU, LOGL_ERROR, "Received %zu bytes on PCU Socket, but primitive "
+			     "container size is %zu, discarding\n", prim_len, exp_len);
+			return -EINVAL;
+		}
+		rc = abis_osmo_pcu_tx_container(bts, &pcu_prim->u.container);
+		break;
 	default:
 		LOGP(DPCU, LOGL_ERROR, "Received unknown PCU msg type %d\n",
 			msg_type);
@@ -905,14 +964,15 @@ struct pcu_sock_state {
 	struct llist_head upqueue;	/* queue for sending messages */
 };
 
-static int pcu_sock_send(struct gsm_network *net, struct msgb *msg)
+int pcu_sock_send(struct gsm_network *net, struct msgb *msg)
 {
 	struct pcu_sock_state *state = net->pcu_state;
 	struct osmo_fd *conn_bfd;
 	struct gsm_pcu_if *pcu_prim = (struct gsm_pcu_if *) msg->data;
 
 	if (!state) {
-		if (pcu_prim->msg_type != PCU_IF_MSG_TIME_IND)
+		if (pcu_prim->msg_type != PCU_IF_MSG_TIME_IND &&
+		    pcu_prim->msg_type != PCU_IF_MSG_INTERF_IND)
 			LOGP(DPCU, LOGL_INFO, "PCU socket not created, "
 				"dropping message\n");
 		msgb_free(msg);
@@ -920,7 +980,8 @@ static int pcu_sock_send(struct gsm_network *net, struct msgb *msg)
 	}
 	conn_bfd = &state->conn_bfd;
 	if (conn_bfd->fd <= 0) {
-		if (pcu_prim->msg_type != PCU_IF_MSG_TIME_IND)
+		if (pcu_prim->msg_type != PCU_IF_MSG_TIME_IND &&
+		    pcu_prim->msg_type != PCU_IF_MSG_INTERF_IND)
 			LOGP(DPCU, LOGL_NOTICE, "PCU socket not connected, "
 				"dropping message\n");
 		msgb_free(msg);
@@ -994,7 +1055,7 @@ static int pcu_sock_read(struct osmo_fd *bfd)
 	struct msgb *msg;
 	int rc;
 
-	msg = msgb_alloc(sizeof(*pcu_prim), "pcu_sock_rx");
+	msg = msgb_alloc(sizeof(*pcu_prim) + 1000, "pcu_sock_rx");
 	if (!msg)
 		return -ENOMEM;
 
@@ -1012,14 +1073,14 @@ static int pcu_sock_read(struct osmo_fd *bfd)
 		goto close;
 	}
 
-	if (rc < sizeof(*pcu_prim)) {
-		LOGP(DPCU, LOGL_ERROR, "Received %d bytes on PCU Socket, but primitive size "
-		     "is %zu, discarding\n", rc, sizeof(*pcu_prim));
+	if (rc < PCUIF_HDR_SIZE) {
+		LOGP(DPCU, LOGL_ERROR, "Received %d bytes on PCU Socket, but primitive hdr size "
+		     "is %zu, discarding\n", rc, PCUIF_HDR_SIZE);
 		msgb_free(msg);
 		return 0;
 	}
 
-	rc = pcu_rx(state->net, pcu_prim->msg_type, pcu_prim);
+	rc = pcu_rx(state->net, pcu_prim->msg_type, pcu_prim, rc);
 
 	/* as we always synchronously process the message in pcu_rx() and
 	 * its callbacks, we can free the message here. */
