@@ -118,7 +118,9 @@ struct phy_instance *phy_instance_create(struct phy_link *plink, int num)
 
 void phy_instance_link_to_trx(struct phy_instance *pinst, struct gsm_bts_trx *trx)
 {
-	trx->role_bts.l1h = pinst;
+	/* There might already be an associated TRX */
+	OSMO_ASSERT(pinst->trx == NULL)
+	trx->pinst = pinst;
 	pinst->trx = trx;
 }
 
@@ -128,8 +130,8 @@ void phy_instance_destroy(struct phy_instance *pinst)
 	llist_del(&pinst->list);
 
 	/* remove reverse link from TRX */
-	OSMO_ASSERT(pinst->trx->role_bts.l1h == pinst);
-	pinst->trx->role_bts.l1h = NULL;
+	OSMO_ASSERT(pinst->trx->pinst == pinst);
+	pinst->trx->pinst = NULL;
 	pinst->trx = NULL;
 
 	talloc_free(pinst);
@@ -146,7 +148,7 @@ void phy_link_destroy(struct phy_link *plink)
 }
 
 static char name_buf[32];
-const char *phy_link_name(struct phy_link *plink)
+const char *phy_link_name(const struct phy_link *plink)
 {
 	snprintf(name_buf, sizeof(name_buf), "phy%u", plink->num);
 	return name_buf;
@@ -154,11 +156,20 @@ const char *phy_link_name(struct phy_link *plink)
 
 int phy_links_open(void)
 {
+	const struct phy_instance *pinst;
 	struct phy_link *plink;
 
 	llist_for_each_entry(plink, &g_phy_links, list) {
 		int rc;
 
+		/* Warn about dangling PHY instances */
+		llist_for_each_entry(pinst, &plink->instances, list) {
+			if (pinst->trx != NULL)
+				continue;
+			LOGPPHI(pinst, DL1C, LOGL_NOTICE, "This PHY instance is not associated "
+				"with a TRX instance, check the configuration file!\n");
+		}
+
 		rc = bts_model_phy_link_open(plink);
 		if (rc < 0)
 			return rc;
@@ -167,7 +178,7 @@ int phy_links_open(void)
 	return 0;
 }
 
-const char *phy_instance_name(struct phy_instance *pinst)
+const char *phy_instance_name(const struct phy_instance *pinst)
 {
 	snprintf(name_buf, sizeof(name_buf), "phy%u.%u", pinst->phy_link->num,
 		 pinst->num);
