@@ -37,11 +37,12 @@
 #include <osmocom/vty/logging.h>
 #include <osmocom/vty/misc.h>
 #include <osmocom/vty/ports.h>
+#include <osmocom/vty/tdef_vty.h>
 #include <osmocom/core/gsmtap.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/core/sockaddr_str.h>
 #include <osmocom/trau/osmo_ortp.h>
-
+#include <osmocom/core/fsm.h>
 
 #include <osmo-bts/logging.h>
 #include <osmo-bts/gsm_data.h>
@@ -94,14 +95,14 @@ struct phy_instance *vty_get_phy_instance(struct vty *vty, int phy_nr, int inst_
 	struct phy_instance *pinst;
 
 	if (!plink) {
-		vty_out(vty, "Cannot find PHY link number %d%s",
+		vty_out(vty, "%% Cannot find PHY link number %d%s",
 			phy_nr, VTY_NEWLINE);
 		return NULL;
 	}
 
 	pinst = phy_instance_by_num(plink, inst_nr);
 	if (!pinst) {
-		vty_out(vty, "Cannot find PHY instance number %d%s",
+		vty_out(vty, "%% Cannot find PHY instance number %d%s",
 			inst_nr, VTY_NEWLINE);
 		return NULL;
 	}
@@ -229,7 +230,7 @@ DEFUN_ATTR(cfg_bts_trx, cfg_bts_trx_cmd,
 		 */
 		trx = gsm_bts_trx_alloc(bts);
 		if (trx)
-			bts_trx_init(trx);
+			bts_model_trx_init(trx);
 	} else
 		trx = gsm_bts_trx_num(bts, trx_nr);
 
@@ -282,6 +283,7 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 	const struct gsm_bts_trx *trx;
 	const char *sapi_buf;
 	int i;
+	struct bsc_oml_host *bsc_oml_host;
 
 	vty_out(vty, "bts %u%s", bts->nr, VTY_NEWLINE);
 	if (bts->description)
@@ -291,7 +293,8 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		vty_out(vty, " auto-band%s", VTY_NEWLINE);
 	vty_out(vty, " ipa unit-id %u %u%s",
 		bts->ip_access.site_id, bts->ip_access.bts_id, VTY_NEWLINE);
-	vty_out(vty, " oml remote-ip %s%s", bts->bsc_oml_host, VTY_NEWLINE);
+	llist_for_each_entry(bsc_oml_host, &bts->bsc_oml_hosts, list)
+		vty_out(vty, " oml remote-ip %s%s", bsc_oml_host->addr, VTY_NEWLINE);
 	vty_out(vty, " rtp jitter-buffer %u", bts->rtp_jitter_buf_ms);
 	if (bts->rtp_jitter_adaptive)
 		vty_out(vty, " adaptive");
@@ -300,6 +303,8 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		bts->rtp_port_range_end, VTY_NEWLINE);
 	if (bts->rtp_ip_dscp != -1)
 		vty_out(vty, " rtp ip-dscp %i%s", bts->rtp_ip_dscp, VTY_NEWLINE);
+	if (bts->rtp_priority != -1)
+		vty_out(vty, " rtp socket-priority %i%s", bts->rtp_priority, VTY_NEWLINE);
 	vty_out(vty, " paging queue-size %u%s", paging_get_queue_max(bts->paging_state),
 		VTY_NEWLINE);
 	vty_out(vty, " paging lifetime %u%s", paging_get_lifetime(bts->paging_state),
@@ -365,6 +370,8 @@ static void config_write_bts_single(struct vty *vty, const struct gsm_bts *bts)
 		vty_out(vty, "  ms-power-control %s%s",
 			trx->ms_pwr_ctl_soft ? "osmo" : "dsp",
 			VTY_NEWLINE);
+		vty_out(vty, "  ta-control interval %u%s",
+			trx->ta_ctrl_interval, VTY_NEWLINE);
 		vty_out(vty, "  phy %u instance %u%s", pinst->phy_link->num,
 			pinst->num, VTY_NEWLINE);
 
@@ -380,6 +387,7 @@ static int config_write_bts(struct vty *vty)
 	llist_for_each_entry(bts, &net->bts_list, list)
 		config_write_bts_single(vty, bts);
 
+	osmo_tdef_vty_groups_write(vty, "");
 	return CMD_SUCCESS;
 }
 
@@ -505,11 +513,51 @@ DEFUN(cfg_bts_oml_ip,
       "OML Parameters\n" "OML IP Address\n" "OML IP Address\n")
 {
 	struct gsm_bts *bts = vty->index;
+	struct bsc_oml_host *bsc_oml_host;
 
-	if (bts->bsc_oml_host)
-		talloc_free(bts->bsc_oml_host);
+	/* stop when the address is already in the list */
+	llist_for_each_entry(bsc_oml_host, &bts->bsc_oml_hosts, list) {
+		if (strcmp(argv[0], bsc_oml_host->addr) == 0) {
+			vty_out(vty, "%% duplicate BSC (A-Bis/OML) ip address configured ('%s')%s",
+				argv[0], VTY_NEWLINE);
+			return CMD_WARNING;
+		}
+	}
 
-	bts->bsc_oml_host = talloc_strdup(bts, argv[0]);
+	bsc_oml_host = talloc_zero(bts, struct bsc_oml_host);
+	OSMO_ASSERT(bsc_oml_host);
+	bsc_oml_host->addr = talloc_strdup(bsc_oml_host, argv[0]);
+	OSMO_ASSERT(bsc_oml_host->addr);
+	llist_add_tail(&bsc_oml_host->list, &bts->bsc_oml_hosts);
+
+	return CMD_SUCCESS;
+}
+
+DEFUN(cfg_bts_no_oml_ip,
+      cfg_bts_no_oml_ip_cmd,
+      "no oml remote-ip A.B.C.D",
+      NO_STR "OML Parameters\n" "OML IP Address\n" "OML IP Address\n")
+{
+	struct gsm_bts *bts = vty->index;
+	struct bsc_oml_host *bsc_oml_host;
+	struct bsc_oml_host *bsc_oml_host_del = NULL;
+
+	llist_for_each_entry(bsc_oml_host, &bts->bsc_oml_hosts, list) {
+		if (strcmp(argv[0], bsc_oml_host->addr) == 0)
+			bsc_oml_host_del = bsc_oml_host;
+	}
+
+	if (bsc_oml_host_del) {
+		if (bts->abis_link_fi) {
+			osmo_fsm_inst_dispatch(bts->abis_link_fi, ABIS_LINK_EV_VTY_RM_ADDR, bsc_oml_host_del);
+			llist_del(&bsc_oml_host_del->list);
+			talloc_free(bsc_oml_host_del);
+		}
+	} else {
+		vty_out(vty, "%% no such BSC (A-Bis/OML) ip address configured ('%s')%s",
+			argv[0], VTY_NEWLINE);
+		return CMD_WARNING;
+	}
 
 	return CMD_SUCCESS;
 }
@@ -557,19 +605,19 @@ DEFUN(cfg_bts_rtp_port_range,
 	end = atoi(argv[1]);
 
 	if (end < start) {
-		vty_out(vty, "range end port (%u) must be greater than the range start port (%u)!%s",
+		vty_out(vty, "%% Range end port (%u) must be greater than the range start port (%u)!%s",
 			end, start, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
 	if (start & 1) {
-		vty_out(vty, "range must begin at an even port number! (%u not even)%s",
+		vty_out(vty, "%% Range must begin at an even port number! (%u is odd)%s",
 			start, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
 	if ((end & 1) == 0) {
-		vty_out(vty, "range must end at an odd port number! (%u not odd)%s",
+		vty_out(vty, "%% Range must end at an odd port number! (%u is even)%s",
 			end, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
@@ -595,6 +643,21 @@ DEFUN_USRATTR(cfg_bts_rtp_ip_dscp,
 	return CMD_SUCCESS;
 }
 
+DEFUN_USRATTR(cfg_bts_rtp_priority,
+	      cfg_bts_rtp_priority_cmd,
+	      X(BTS_VTY_ATTR_NEW_LCHAN),
+	      "rtp socket-priority <0-255>",
+	      RTP_STR "Specify socket priority for RTP/IP packets\n"
+	      "The socket priority value (> 6 requires CAP_NET_ADMIN)\n")
+{
+	struct gsm_bts *bts = vty->index;
+	int prio = atoi(argv[0]);
+
+	bts->rtp_priority = prio;
+
+	return CMD_SUCCESS;
+}
+
 #define PAG_STR "Paging related parameters\n"
 
 DEFUN_ATTR(cfg_bts_paging_queue_size,
@@ -935,11 +998,11 @@ DEFUN(cfg_trx_ms_power_control, cfg_trx_ms_power_control_cmd,
 		 * has been ripped out in favour of the common implementation. Configuration files
 		 * may still contain 'dsp', so let's be tolerant and override 'dsp' by 'osmo'. */
 		if (trx->bts->variant == BTS_OSMO_TRX && vty->type == VTY_FILE) {
-			vty_out(vty, "BTS model 'osmo-bts-trx' has no DSP/HW MS Power Control support, "
+			vty_out(vty, "%% BTS model 'osmo-bts-trx' has no DSP/HW MS Power Control support, "
 				     "consider updating your configuration file!%s", VTY_NEWLINE);
 			soft = true; /* override */
 		} else {
-			vty_out(vty, "This BTS model has no DSP/HW MS Power Control support%s", VTY_NEWLINE);
+			vty_out(vty, "%% This BTS model has no DSP/HW MS Power Control support%s", VTY_NEWLINE);
 			return CMD_WARNING;
 		}
 	}
@@ -948,6 +1011,19 @@ DEFUN(cfg_trx_ms_power_control, cfg_trx_ms_power_control_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(cfg_ta_ctrl_interval, cfg_ta_ctrl_interval_cmd,
+	"ta-control interval <0-31>",
+	"Timing Advance Control Parameters\n"
+	"Set TA control loop interval\n"
+	"As in P_CON_INTERVAL, in units of 2 SACCH periods (0.96 seconds) (default=0, every SACCH block)\n")
+{
+	struct gsm_bts_trx *trx = vty->index;
+
+	trx->ta_ctrl_interval = atoi(argv[0]);
+
+	return CMD_SUCCESS;
+}
+
 DEFUN(cfg_trx_phy, cfg_trx_phy_cmd,
 	"phy <0-255> instance <0-255>",
 	"Configure PHY Link+Instance for this TRX\n"
@@ -958,20 +1034,25 @@ DEFUN(cfg_trx_phy, cfg_trx_phy_cmd,
 	struct phy_instance *pinst;
 
 	if (!plink) {
-		vty_out(vty, "phy%s does not exist%s",
+		vty_out(vty, "%% phy%s does not exist%s",
 			argv[0], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
 	pinst = phy_instance_by_num(plink, atoi(argv[1]));
 	if (!pinst) {
-		vty_out(vty, "phy%s instance %s does not exit%s",
+		vty_out(vty, "%% phy%s instance %s does not exit%s",
 			argv[0], argv[1], VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
-	trx->role_bts.l1h = pinst;
-	pinst->trx = trx;
+	if (pinst->trx != NULL) {
+		vty_out(vty, "%% phy%s instance %s is already bound to %s%s",
+			argv[0], argv[1], gsm_trx_name(pinst->trx), VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	phy_instance_link_to_trx(pinst, trx);
 
 	return CMD_SUCCESS;
 }
@@ -998,7 +1079,7 @@ static void bts_dump_vty_features(struct vty *vty, const struct gsm_bts *bts)
 	for (i = 0, no_features = true; i < _NUM_BTS_FEAT; i++) {
 		if (osmo_bts_has_feature(bts->features, i)) {
 			vty_out(vty, "    %03u ", i);
-			vty_out(vty, "%-40s%s", osmo_bts_feature_name(i), VTY_NEWLINE);
+			vty_out(vty, "%-40s%s", osmo_bts_features_desc(i), VTY_NEWLINE);
 			no_features = false;
 		}
 	}
@@ -1084,6 +1165,10 @@ static void bts_dump_vty(struct vty *vty, const struct gsm_bts *bts)
 		vty_out(vty, "  Radio Link Timeout (OVERRIDE): %s%s",
 			stringify_radio_link_timeout(bts->radio_link_timeout.current), VTY_NEWLINE);
 	}
+	if (bts->c0_power_red_db > 0) {
+		vty_out(vty, "  BCCH carrier power reduction: %u dB%s",
+			bts->c0_power_red_db, VTY_NEWLINE);
+	}
 
 	llist_for_each_entry(trx, &bts->trx_list, list) {
 		const struct phy_instance *pinst = trx_phy_instance(trx);
@@ -1242,6 +1327,39 @@ DEFUN_HIDDEN(radio_link_timeout, radio_link_timeout_cmd, "bts <0-0> radio-link-t
 	return CMD_SUCCESS;
 }
 
+DEFUN(bts_c0_power_red,
+      bts_c0_power_red_cmd,
+      "bts <0-255> c0-power-red <0-6>",
+      "BTS Specific Commands\n" BTS_NR_STR
+      "BCCH carrier power reduction operation\n"
+      "Power reduction value (in dB, even numbers only)\n")
+{
+	struct gsm_network *net = gsmnet_from_vty(vty);
+	const int bts_nr = atoi(argv[0]);
+	const int red = atoi(argv[1]);
+	struct gsm_bts *bts;
+
+	bts = gsm_bts_num(net, atoi(argv[0]));
+	if (bts == NULL) {
+		vty_out(vty, "%% No such BTS (%d)%s", bts_nr, VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (red % 2 != 0) {
+		vty_out(vty, "%% Incorrect BCCH power reduction value, "
+			"an even number is expected%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	if (bts_set_c0_pwr_red(bts, red) != 0) {
+		vty_out(vty, "%% BCCH carrier power reduction operation mode "
+			"is not supported for BTS (%d)%s", bts_nr, VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+
+	return CMD_SUCCESS;
+}
+
 /* TODO: generalize and move indention handling to libosmocore */
 #define cfg_out(vty, fmt, args...) \
 	vty_out(vty, "%*s" fmt, indent, "", ##args)
@@ -1288,7 +1406,7 @@ static void dump_dpc_meas_params(struct vty *vty, const unsigned int indent,
 }
 
 static void dump_dpc_params(struct vty *vty, const unsigned int indent,
-			    const struct gsm_power_ctrl_params *cp)
+			    const struct gsm_power_ctrl_params *cp, bool uplink)
 {
 	cfg_out(vty, "Power control interval: %u ms (every %u SACCH block(s))%s",
 		cp->ctrl_interval ? cp->ctrl_interval * 2 * 480 : 480,
@@ -1305,6 +1423,26 @@ static void dump_dpc_params(struct vty *vty, const unsigned int indent,
 
 	cfg_out(vty, "RxQual measurement processing:%s", VTY_NEWLINE);
 	dump_dpc_meas_params(vty, indent + 2, &cp->rxqual_meas, "RXQUAL", 3);
+
+	if (uplink) {
+		cfg_out(vty, "C/I measurement processing (FR/EFR):%s", VTY_NEWLINE);
+		dump_dpc_meas_params(vty, indent + 2, &cp->ci_fr_meas, "CI_FR", 0);
+
+		cfg_out(vty, "C/I measurement processing (HR):%s", VTY_NEWLINE);
+		dump_dpc_meas_params(vty, indent + 2, &cp->ci_hr_meas, "CI_HR", 0);
+
+		cfg_out(vty, "C/I measurement processing (AMR-FR):%s", VTY_NEWLINE);
+		dump_dpc_meas_params(vty, indent + 2, &cp->ci_amr_fr_meas, "CI_AMR_FR", 0);
+
+		cfg_out(vty, "C/I measurement processing (AMR-HR):%s", VTY_NEWLINE);
+		dump_dpc_meas_params(vty, indent + 2, &cp->ci_amr_hr_meas, "CI_AMR_HR", 0);
+
+		cfg_out(vty, "C/I measurement processing (SDCCH):%s", VTY_NEWLINE);
+		dump_dpc_meas_params(vty, indent + 2, &cp->ci_sdcch_meas, "CI_SDCCH", 0);
+
+		cfg_out(vty, "C/I measurement processing (GPRS):%s", VTY_NEWLINE);
+		dump_dpc_meas_params(vty, indent + 2, &cp->ci_gprs_meas, "CI_GPRS", 0);
+	}
 }
 
 static void trx_dump_vty(struct vty *vty, const struct gsm_bts_trx *trx)
@@ -1322,13 +1460,13 @@ static void trx_dump_vty(struct vty *vty, const struct gsm_bts_trx *trx)
 		trx->bs_dpc_params == &trx->bts->bs_dpc_params ?
 			"fall-back" : "from BSC",
 		VTY_NEWLINE);
-	dump_dpc_params(vty, 4, trx->bs_dpc_params);
+	dump_dpc_params(vty, 4, trx->bs_dpc_params, false);
 
 	vty_out(vty, "  MS Power control parameters (%s):%s",
 		trx->ms_dpc_params == &trx->bts->ms_dpc_params ?
 			"fall-back" : "from BSC",
 		VTY_NEWLINE);
-	dump_dpc_params(vty, 4, trx->ms_dpc_params);
+	dump_dpc_params(vty, 4, trx->ms_dpc_params, true);
 
 	vty_out(vty, "  NM State: ");
 	net_dump_nmstate(vty, &trx->mo.nm_state);
@@ -1392,7 +1530,7 @@ static void ts_dump_vty(struct vty *vty, const struct gsm_bts_trx_ts *ts)
 {
 	vty_out(vty, "BTS %u, TRX %u, Timeslot %u, phys cfg %s, TSC %u",
 		ts->trx->bts->nr, ts->trx->nr, ts->nr,
-		gsm_pchan_name(ts->pchan), gsm_ts_tsc(ts));
+		gsm_pchan_name(ts->pchan), ts->tsc);
 	if (ts->pchan == GSM_PCHAN_TCH_F_PDCH)
 		vty_out(vty, " (%s mode)",
 			ts->flags & TS_F_PDCH_ACTIVE ? "PDCH" : "TCH/F");
@@ -1482,7 +1620,7 @@ DEFUN(show_ts,
 static void vty_out_dyn_ts_status(struct vty *vty, const struct gsm_bts_trx_ts *ts)
 {
 	switch (ts->pchan) {
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		if (ts->dyn.pchan_is == ts->dyn.pchan_want)
 			vty_out(vty, " as %s",
 				gsm_pchan_name(ts->dyn.pchan_is));
@@ -1535,7 +1673,7 @@ static void lchan_bs_power_ctrl_state_dump(struct vty *vty, unsigned int indent,
 		return;
 
 	cfg_out(vty, "Power Control parameters:%s", VTY_NEWLINE);
-	dump_dpc_params(vty, indent + 2, st->dpc_params);
+	dump_dpc_params(vty, indent + 2, st->dpc_params, false);
 }
 
 static void lchan_ms_power_ctrl_state_dump(struct vty *vty, unsigned int indent,
@@ -1561,19 +1699,105 @@ static void lchan_ms_power_ctrl_state_dump(struct vty *vty, unsigned int indent,
 		return;
 
 	cfg_out(vty, "Power Control parameters:%s", VTY_NEWLINE);
-	dump_dpc_params(vty, indent + 2, st->dpc_params);
+	dump_dpc_params(vty, indent + 2, st->dpc_params, true);
+}
+
+static void lchan_acch_rep_state_dump(struct vty *vty, unsigned int indent,
+				      const struct gsm_lchan *lchan)
+{
+	cfg_out(vty, "ACCH repetition:%s", VTY_NEWLINE);
+	indent += 2;
+	if (lchan->rep_acch_cap.rxqual)
+		cfg_out(vty, "Enable RXQUAL threshold: %u%s",
+			lchan->rep_acch_cap.rxqual, VTY_NEWLINE);
+	else
+		cfg_out(vty, "Enable RXQUAL threshold: (none, alway on)%s",
+			VTY_NEWLINE);
+
+	cfg_out(vty, "DL-FACCH:%s", VTY_NEWLINE);
+	indent += 2;
+	if (lchan->rep_acch_cap.dl_facch_all)
+		cfg_out(vty, "retramsit all LAPDM block types%s", VTY_NEWLINE);
+	else if (lchan->rep_acch_cap.dl_facch_cmd)
+		cfg_out(vty, "retramsit only LAPDM command blocks%s",
+			VTY_NEWLINE);
+	else
+		cfg_out(vty, "no retransmission (disabled)%s", VTY_NEWLINE);
+	if (lchan->rep_acch.dl_facch_active)
+		cfg_out(vty, "retransmission currently active%s", VTY_NEWLINE);
+	else
+		cfg_out(vty, "retransmission currently inactive%s",
+			VTY_NEWLINE);
+	indent -= 2;
+
+	cfg_out(vty, "DL-SACCH:%s", VTY_NEWLINE);
+	indent += 2;
+	if (lchan->rep_acch_cap.ul_sacch)
+		cfg_out(vty, "retramsit all SACCH blocks for SAPI=0%s",
+			VTY_NEWLINE);
+	else
+		cfg_out(vty, "no retransmission (disabled)%s", VTY_NEWLINE);
+	if (lchan->rep_acch.dl_sacch_active)
+		cfg_out(vty, "retransmission currently active%s", VTY_NEWLINE);
+	else
+		cfg_out(vty, "retransmission currently inactive%s",
+			VTY_NEWLINE);
+	indent -= 2;
+
+	cfg_out(vty, "UL-SACCH:%s", VTY_NEWLINE);
+	indent += 2;
+	if (lchan->rep_acch_cap.dl_sacch)
+		cfg_out(vty, "retramsit all SACCH blocks for SAPI=0%s",
+			VTY_NEWLINE);
+	else
+		cfg_out(vty, "no retransmission (disabled)%s", VTY_NEWLINE);
+	if (lchan->rep_acch.ul_sacch_active)
+		cfg_out(vty, "retransmission currently active%s", VTY_NEWLINE);
+	else
+		cfg_out(vty, "retransmission currently inactive%s",
+			VTY_NEWLINE);
+	indent -= 2;
+}
+
+static void lchan_acch_top_state_dump(struct vty *vty, unsigned int indent,
+				      const struct gsm_lchan *lchan)
+{
+	if (lchan->top_acch_cap.overpower_db == 0)
+		return;
+
+	cfg_out(vty, "Temporary ACCH overpower:%s", VTY_NEWLINE);
+	indent += 2;
+
+	cfg_out(vty, "Overpower value: %u dB%s",
+		lchan->top_acch_cap.overpower_db, VTY_NEWLINE);
+
+	cfg_out(vty, "SACCH overpower: %sabled%s",
+		lchan->top_acch_cap.sacch_enable ? "en" : "dis",
+		VTY_NEWLINE);
+	cfg_out(vty, "FACCH overpower: %sabled%s",
+		lchan->top_acch_cap.facch_enable ? "en" : "dis",
+		VTY_NEWLINE);
+
+	if (lchan->top_acch_cap.rxqual == 0) {
+		cfg_out(vty, "RxQual threshold: disabled "
+			"(overpower is always on)%s", VTY_NEWLINE);
+	} else {
+		cfg_out(vty, "RxQual threshold: %u%s",
+			lchan->top_acch_cap.rxqual, VTY_NEWLINE);
+	}
 }
 
 static void lchan_dump_full_vty(struct vty *vty, const struct gsm_lchan *lchan)
 {
 	struct in_addr ia;
 
-	vty_out(vty, "BTS %u, TRX %u, Timeslot %u, Lchan %u: Type %s%s",
+	vty_out(vty, "BTS %u, TRX %u, Timeslot %u (%s), Lchan %u: Type %s%s",
 		lchan->ts->trx->bts->nr, lchan->ts->trx->nr, lchan->ts->nr,
+		lchan->ts->vamos.is_shadow ? "shadow" : "primary",
 		lchan->nr, gsm_lchant_name(lchan->type), VTY_NEWLINE);
 	/* show dyn TS details, if applicable */
 	switch (lchan->ts->pchan) {
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		vty_out(vty, "  Osmocom Dyn TS:");
 		vty_out_dyn_ts_status(vty, lchan->ts);
 		vty_out(vty, VTY_NEWLINE);
@@ -1634,17 +1858,30 @@ static void lchan_dump_full_vty(struct vty *vty, const struct gsm_lchan *lchan)
 		vty_out(vty, "  RTP/PDCH Loopback Enabled%s", VTY_NEWLINE);
 	vty_out(vty, "  Radio Link Failure Counter 'S': %d%s", lchan->s, VTY_NEWLINE);
 
+	/* Interference levels */
+	if (lchan->meas.interf_meas_avg_dbm != 0) {
+		vty_out(vty, "  Interference: %d dBm (band %d)%s",
+			lchan->meas.interf_meas_avg_dbm,
+			lchan->meas.interf_band,
+			VTY_NEWLINE);
+	}
+
 	/* BS/MS Power Control state and parameters */
 	lchan_bs_power_ctrl_state_dump(vty, 2, lchan);
 	lchan_ms_power_ctrl_state_dump(vty, 2, lchan);
+
+	/* ACCH repetition / overpower state */
+	lchan_acch_rep_state_dump(vty, 2, lchan);
+	lchan_acch_top_state_dump(vty, 2, lchan);
 }
 
 static void lchan_dump_short_vty(struct vty *vty, const struct gsm_lchan *lchan)
 {
 	const struct gsm_meas_rep_unidir *mru = &lchan->meas.ul_res;
 
-	vty_out(vty, "BTS %u, TRX %u, Timeslot %u %s",
+	vty_out(vty, "BTS %u, TRX %u, Timeslot %u (%s) %s",
 		lchan->ts->trx->bts->nr, lchan->ts->trx->nr, lchan->ts->nr,
+		lchan->ts->vamos.is_shadow ? "shadow" : "primary",
 		gsm_pchan_name(lchan->ts->pchan));
 	vty_out_dyn_ts_status(vty, lchan->ts);
 	vty_out(vty, ", Lchan %u, Type %s, State %s - "
@@ -1685,6 +1922,8 @@ static int dump_lchan_trx(const struct gsm_bts_trx *trx, struct vty *vty,
 	for (ts_nr = 0; ts_nr < TRX_NR_TS; ts_nr++) {
 		const struct gsm_bts_trx_ts *ts = &trx->ts[ts_nr];
 		dump_lchan_trx_ts(ts, vty, dump_cb);
+		if (ts->vamos.peer != NULL) /* VAMOS: shadow timeslot */
+			dump_lchan_trx_ts(ts->vamos.peer, vty, dump_cb);
 	}
 
 	return CMD_SUCCESS;
@@ -1707,9 +1946,9 @@ static int lchan_summary(struct vty *vty, int argc, const char **argv,
 			 void (*dump_cb)(struct vty *, const struct gsm_lchan *))
 {
 	const struct gsm_network *net = gsmnet_from_vty(vty);
-	const struct gsm_bts *bts;
-	const struct gsm_bts_trx *trx;
-	const struct gsm_bts_trx_ts *ts;
+	const struct gsm_bts *bts = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
+	const struct gsm_bts_trx *trx = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
+	const struct gsm_bts_trx_ts *ts = NULL; /* initialize to avoid uninitialized false warnings on some gcc versions (11.1.0) */
 	const struct gsm_lchan *lchan;
 	int bts_nr, trx_nr, ts_nr, lchan_nr;
 
@@ -1795,7 +2034,8 @@ static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
 	int bts_nr = atoi(argv[idx+0]);
 	int trx_nr = atoi(argv[idx+1]);
 	int ts_nr = atoi(argv[idx+2]);
-	int lchan_nr = atoi(argv[idx+3]);
+	bool shadow = argv[idx+3][0] == 's';
+	int lchan_nr = atoi(argv[idx+4]);
 	struct gsm_bts *bts;
 	struct gsm_bts_trx *trx;
 	struct gsm_bts_trx_ts *ts;
@@ -1811,6 +2051,11 @@ static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
 	if (ts_nr >= ARRAY_SIZE(trx->ts))
 		return NULL;
 	ts = &trx->ts[ts_nr];
+	if (shadow) { /* VAMOS shadow */
+		if (ts->vamos.peer == NULL)
+			return NULL;
+		ts = ts->vamos.peer;
+	}
 
 	if (lchan_nr >= ARRAY_SIZE(ts->lchan))
 		return NULL;
@@ -1819,7 +2064,7 @@ static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
 }
 
 #define BTS_T_T_L_CMD \
-	"bts <0-0> trx <0-255> ts <0-7> lchan <0-7>"
+	"bts <0-0> trx <0-255> ts <0-7> (lchan|shadow-lchan) <0-7>"
 #define BTS_T_T_L_STR			\
 	"BTS related commands\n"	\
 	"BTS number\n"			\
@@ -1827,7 +2072,8 @@ static struct gsm_lchan *resolve_lchan(const struct gsm_network *net,
 	"TRX number\n"			\
 	"timeslot related commands\n"	\
 	"timeslot number\n"		\
-	"logical channel commands\n"	\
+	"Primary logical channel commands\n"	\
+	"Shadow logical channel commands\n"	\
 	"logical channel number\n"
 
 DEFUN(cfg_bts_gsmtap_remote_host,
@@ -1963,7 +2209,7 @@ DEFUN_ATTR(cfg_phy_inst, cfg_phy_inst_cmd,
 	if (!pinst) {
 		pinst = phy_instance_create(plink, inst_nr);
 		if (!pinst) {
-			vty_out(vty, "Unable to create phy%u instance %u%s",
+			vty_out(vty, "%% Unable to create phy%u instance %u%s",
 				plink->num, inst_nr, VTY_NEWLINE);
 			return CMD_WARNING;
 		}
@@ -1986,7 +2232,7 @@ DEFUN(cfg_phy_no_inst, cfg_phy_no_inst_cmd,
 
 	pinst = phy_instance_by_num(plink, inst_nr);
 	if (!pinst) {
-		vty_out(vty, "No such instance %u%s", inst_nr, VTY_NEWLINE);
+		vty_out(vty, "%% No such instance %u%s", inst_nr, VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -2006,7 +2252,7 @@ DEFUN(cfg_phy_type, cfg_phy_type_cmd,
 	struct phy_link *plink = vty->index;
 
 	if (plink->state != PHY_LINK_SHUTDOWN) {
-		vty_out(vty, "Cannot change type of active PHY%s", VTY_NEWLINE);
+		vty_out(vty, "%% Cannot change type of active PHY%s", VTY_NEWLINE);
 		return CMD_WARNING;
 	}
 
@@ -2248,12 +2494,17 @@ int bts_vty_init(void *ctx)
 	install_node(&bts_node, config_write_bts);
 	install_element(CONFIG_NODE, &cfg_bts_cmd);
 	install_element(CONFIG_NODE, &cfg_vty_telnet_port_cmd);
+
+	osmo_tdef_vty_groups_init(CONFIG_NODE, bts_tdef_groups);
+
 	install_element(BTS_NODE, &cfg_bts_unit_id_cmd);
 	install_element(BTS_NODE, &cfg_bts_oml_ip_cmd);
+	install_element(BTS_NODE, &cfg_bts_no_oml_ip_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_bind_ip_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_jitbuf_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_port_range_cmd);
 	install_element(BTS_NODE, &cfg_bts_rtp_ip_dscp_cmd);
+	install_element(BTS_NODE, &cfg_bts_rtp_priority_cmd);
 	install_element(BTS_NODE, &cfg_bts_band_cmd);
 	install_element(BTS_NODE, &cfg_description_cmd);
 	install_element(BTS_NODE, &cfg_no_description_cmd);
@@ -2290,6 +2541,7 @@ int bts_vty_init(void *ctx)
 	install_element(TRX_NODE, &cfg_trx_pr_step_size_cmd);
 	install_element(TRX_NODE, &cfg_trx_pr_step_interval_cmd);
 	install_element(TRX_NODE, &cfg_trx_ms_power_control_cmd);
+	install_element(TRX_NODE, &cfg_ta_ctrl_interval_cmd);
 	install_element(TRX_NODE, &cfg_trx_phy_cmd);
 
 	install_element(ENABLE_NODE, &bts_t_t_l_jitter_buf_cmd);
@@ -2299,6 +2551,7 @@ int bts_vty_init(void *ctx)
 	install_element(ENABLE_NODE, &bts_t_t_l_power_ctrl_current_max_cmd);
 	install_element(ENABLE_NODE, &test_send_failure_event_report_cmd);
 	install_element(ENABLE_NODE, &radio_link_timeout_cmd);
+	install_element(ENABLE_NODE, &bts_c0_power_red_cmd);
 
 	install_element(CONFIG_NODE, &cfg_phy_cmd);
 	install_node(&phy_node, config_write_phy);
