@@ -87,31 +87,8 @@ struct trx_l1h *trx_l1h_alloc(void *tall_ctx, struct phy_instance *pinst)
 	return l1h;
 }
 
-static void check_transceiver_availability_trx(struct trx_l1h *l1h, int avail)
-{
-	struct phy_instance *pinst = l1h->phy_inst;
-	struct gsm_bts_trx *trx = pinst->trx;
-
-	/* HACK, we should change state when we receive first clock from
-	 * transceiver */
-	if (avail) {
-		/* signal availability */
-		if (!pinst->u.osmotrx.sw_act_reported) {
-			osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_SW_ACT, NULL);
-			osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_SW_ACT, NULL);
-			pinst->u.osmotrx.sw_act_reported = true;
-		}
-	} else {
-		osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_DISABLE, NULL);
-		osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_DISABLE, NULL);
-	}
-}
-
 int bts_model_lchan_deactivate(struct gsm_lchan *lchan)
 {
-	struct phy_instance *pinst = trx_phy_instance(lchan->ts->trx);
-	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
-
 	if (lchan->rel_act_kind == LCHAN_REL_ACT_REACT) {
 		lchan->rel_act_kind = LCHAN_REL_ACT_RSL;
 		/* FIXME: perform whatever is needed (if any) to set proper PCH/AGCH allocation according to
@@ -121,14 +98,12 @@ int bts_model_lchan_deactivate(struct gsm_lchan *lchan)
 	/* set lchan inactive */
 	lchan_set_state(lchan, LCHAN_S_NONE);
 
-	return trx_sched_set_lchan(&l1h->l1s, gsm_lchan2chan_nr(lchan), LID_DEDIC, false);
+	return trx_sched_set_lchan(lchan, gsm_lchan2chan_nr(lchan), LID_DEDIC, false);
 }
 
 int bts_model_lchan_deactivate_sacch(struct gsm_lchan *lchan)
 {
-	struct phy_instance *pinst = trx_phy_instance(lchan->ts->trx);
-	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
-	return trx_sched_set_lchan(&l1h->l1s, gsm_lchan2chan_nr(lchan), LID_SACCH, false);
+	return trx_sched_set_lchan(lchan, gsm_lchan2chan_nr(lchan), LID_SACCH, false);
 }
 
 int l1if_trx_start_power_ramp(struct gsm_bts_trx *trx, ramp_compl_cb_t ramp_compl_cb)
@@ -190,10 +165,6 @@ static int trx_init(struct gsm_bts_trx *trx)
 	if (rc != 0)
 		return osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OPSTART_NACK,
 					      (void*)(intptr_t)NM_NACK_CANT_PERFORM);
-
-	if (trx == trx->bts->c0)
-		lchan_init_lapdm(&trx->ts[0].lchan[CCCH_LCHAN]);
-
 	/* Send OPSTART ack */
 	return osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_OPSTART_ACK, NULL);
 }
@@ -216,13 +187,15 @@ void bts_model_trx_close(struct gsm_bts_trx *trx)
 	osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CLOSE, NULL);
 
 	/* Set to Operational State: Disabled */
-	check_transceiver_availability_trx(l1h, 0);
+	osmo_fsm_inst_dispatch(trx->mo.fi, NM_EV_DISABLE, NULL);
+	osmo_fsm_inst_dispatch(trx->bb_transc.mo.fi, NM_EV_DISABLE, NULL);
 }
 
-/* on RSL failure, deactivate transceiver */
 void bts_model_abis_close(struct gsm_bts *bts)
 {
-	bts_shutdown(bts, "Abis close");
+	/* Go into shutdown state deactivating transceivers until Abis link
+	 * becomes up again */
+	bts_shutdown_ext(bts, "Abis close", false);
 }
 
 int bts_model_adjst_ms_pwr(struct gsm_lchan *lchan)
@@ -235,15 +208,19 @@ int bts_model_adjst_ms_pwr(struct gsm_lchan *lchan)
 /* set bts attributes */
 static uint8_t trx_set_bts(struct gsm_bts *bts, struct tlv_parsed *new_attr)
 {
-	struct gsm_bts_trx *trx;
+	struct phy_instance *pinst = trx_phy_instance(bts->c0);
+	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 	uint8_t bsic = bts->bsic;
+	struct gsm_bts_trx *trx;
+
+	/* ARFCN for C0 is assigned during Set BTS Attr, see oml.c */
+	osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CFG_ARFCN, (void *)(intptr_t)pinst->trx->arfcn);
 
 	llist_for_each_entry(trx, &bts->trx_list, list) {
-		struct phy_instance *pinst = trx_phy_instance(trx);
-		struct phy_link *plink = pinst->phy_link;
-		struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
+		pinst = trx_phy_instance(trx);
+		l1h = pinst->u.osmotrx.hdl;
+
 		osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CFG_BSIC, (void*)(intptr_t)bsic);
-		check_transceiver_availability_trx(l1h, phy_link_state_get(plink) != PHY_LINK_SHUTDOWN);
 	}
 
 	return 0;
@@ -257,7 +234,9 @@ static uint8_t trx_set_trx(struct gsm_bts_trx *trx)
 	struct phy_link *plink = pinst->phy_link;
 	uint16_t arfcn = trx->arfcn;
 
-	osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CFG_ARFCN, (void*)(intptr_t)arfcn);
+	/* ARFCN for C0 is assigned during Set BTS Attr, see oml.c */
+	if (trx != trx->bts->c0)
+		osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CFG_ARFCN, (void *)(intptr_t)arfcn);
 
 	/* Begin to ramp up the power if power reduction is set by OML and TRX
 	   is already running. Otherwise skip, power ramping will be started
@@ -276,15 +255,9 @@ static uint8_t trx_set_ts_as_pchan(struct gsm_bts_trx_ts *ts,
 	struct phy_instance *pinst = trx_phy_instance(ts->trx);
 	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 	uint8_t tn = ts->nr;
-	uint16_t tsc = ts->tsc;
 	uint8_t slottype;
 	int rc;
 
-	/* all TSC of all timeslots must be equal, because transceiver only
-	 * supports one TSC per TRX */
-
-	osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CFG_TSC, (void*)(intptr_t)tsc);
-
 	/* ignore disabled slots */
 	if (!(l1h->config.slotmask & (1 << tn)))
 		return NM_NACK_RES_NOTAVAIL;
@@ -292,8 +265,8 @@ static uint8_t trx_set_ts_as_pchan(struct gsm_bts_trx_ts *ts,
 	/* set physical channel. For dynamic timeslots, the caller should have
 	 * decided on a more specific PCHAN type already. */
 	OSMO_ASSERT(pchan != GSM_PCHAN_TCH_F_PDCH);
-	OSMO_ASSERT(pchan != GSM_PCHAN_TCH_F_TCH_H_PDCH);
-	rc = trx_sched_set_pchan(&l1h->l1s, tn, pchan);
+	OSMO_ASSERT(pchan != GSM_PCHAN_OSMO_DYN);
+	rc = trx_sched_set_pchan(ts, pchan);
 	if (rc)
 		return NM_NACK_RES_NOTAVAIL;
 
@@ -308,6 +281,13 @@ static uint8_t trx_set_ts_as_pchan(struct gsm_bts_trx_ts *ts,
 
 
 	struct trx_prov_ev_cfg_ts_data data = { .tn = tn, .slottype = slottype };
+	if (ts->tsc_set != 0) {
+		/* On TRXC we use 3GPP compliant numbering, so +1 */
+		data.tsc_set = ts->tsc_set + 1;
+		data.tsc_val = ts->tsc;
+		data.tsc_valid = true;
+	}
+
 	osmo_fsm_inst_dispatch(l1h->provision_fi, TRX_PROV_EV_CFG_TS, &data);
 
 	return 0;
@@ -326,7 +306,7 @@ static uint8_t trx_set_ts(struct gsm_bts_trx_ts *ts)
 		pchan = (ts->flags & TS_F_PDCH_ACTIVE)? GSM_PCHAN_PDCH
 			                              : GSM_PCHAN_TCH_F;
 		break;
-	case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+	case GSM_PCHAN_OSMO_DYN:
 		OSMO_ASSERT(ts->dyn.pchan_is == ts->dyn.pchan_want);
 		pchan = ts->dyn.pchan_is;
 		break;
@@ -344,8 +324,7 @@ static uint8_t trx_set_ts(struct gsm_bts_trx_ts *ts)
  */
 
 /* enable ciphering */
-static int l1if_set_ciphering(struct trx_l1h *l1h, struct gsm_lchan *lchan,
-	uint8_t chan_nr, int downlink)
+static int l1if_set_ciphering(struct gsm_lchan *lchan, uint8_t chan_nr, int downlink)
 {
 	/* ciphering already enabled in both directions */
 	if (lchan->ciph_state == LCHAN_CIPH_RXTX_CONF)
@@ -353,28 +332,22 @@ static int l1if_set_ciphering(struct trx_l1h *l1h, struct gsm_lchan *lchan,
 
 	if (!downlink) {
 		/* set uplink */
-		trx_sched_set_cipher(&l1h->l1s, chan_nr, 0, lchan->encr.alg_id - 1,
-			lchan->encr.key, lchan->encr.key_len);
+		trx_sched_set_cipher(lchan, chan_nr, false);
 		lchan->ciph_state = LCHAN_CIPH_RX_CONF;
 	} else {
 		/* set downlink and also set uplink, if not already */
-		if (lchan->ciph_state != LCHAN_CIPH_RX_CONF) {
-			trx_sched_set_cipher(&l1h->l1s, chan_nr, 0,
-				lchan->encr.alg_id - 1, lchan->encr.key,
-				lchan->encr.key_len);
-		}
-		trx_sched_set_cipher(&l1h->l1s, chan_nr, 1, lchan->encr.alg_id - 1,
-			lchan->encr.key, lchan->encr.key_len);
+		if (lchan->ciph_state != LCHAN_CIPH_RX_CONF)
+			trx_sched_set_cipher(lchan, chan_nr, false);
+		trx_sched_set_cipher(lchan, chan_nr, true);
 		lchan->ciph_state = LCHAN_CIPH_RXTX_CONF;
 	}
 
 	return 0;
 }
 
-static int mph_info_chan_confirm(struct trx_l1h *l1h, uint8_t chan_nr,
-	enum osmo_mph_info_type type, uint8_t cause)
+static int mph_info_chan_confirm(struct gsm_bts_trx *trx, uint8_t chan_nr,
+				 enum osmo_mph_info_type type, uint8_t cause)
 {
-	struct phy_instance *pinst = l1h->phy_inst;
 	struct osmo_phsap_prim l1sap;
 
 	memset(&l1sap, 0, sizeof(l1sap));
@@ -384,7 +357,7 @@ static int mph_info_chan_confirm(struct trx_l1h *l1h, uint8_t chan_nr,
 	l1sap.u.info.u.act_cnf.chan_nr = chan_nr;
 	l1sap.u.info.u.act_cnf.cause = cause;
 
-	return l1sap_up(pinst->trx, &l1sap);
+	return l1sap_up(trx, &l1sap);
 }
 
 int l1if_mph_time_ind(struct gsm_bts *bts, uint32_t fn)
@@ -406,8 +379,6 @@ int l1if_mph_time_ind(struct gsm_bts *bts, uint32_t fn)
 /* primitive from common part */
 int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 {
-	struct phy_instance *pinst = trx_phy_instance(trx);
-	struct trx_l1h *l1h = pinst->u.osmotrx.hdl;
 	struct msgb *msg = l1sap->oph.msg;
 	uint8_t chan_nr;
 	int rc = 0;
@@ -418,21 +389,21 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 		if (!msg)
 			break;
 		/* put data into scheduler's queue */
-		return trx_sched_ph_data_req(&l1h->l1s, l1sap);
+		return trx_sched_ph_data_req(trx, l1sap);
 	case OSMO_PRIM(PRIM_TCH, PRIM_OP_REQUEST):
 		if (!msg)
 			break;
 		/* put data into scheduler's queue */
-		return trx_sched_tch_req(&l1h->l1s, l1sap);
+		return trx_sched_tch_req(trx, l1sap);
 	case OSMO_PRIM(PRIM_MPH_INFO, PRIM_OP_REQUEST):
 		switch (l1sap->u.info.type) {
 		case PRIM_INFO_ACT_CIPH:
 			chan_nr = l1sap->u.info.u.ciph_req.chan_nr;
 			lchan = get_lchan_by_chan_nr(trx, chan_nr);
 			if (l1sap->u.info.u.ciph_req.uplink)
-				l1if_set_ciphering(l1h, lchan, chan_nr, 0);
+				l1if_set_ciphering(lchan, chan_nr, 0);
 			if (l1sap->u.info.u.ciph_req.downlink)
-				l1if_set_ciphering(l1h, lchan, chan_nr, 1);
+				l1if_set_ciphering(lchan, chan_nr, 1);
 			break;
 		case PRIM_INFO_ACTIVATE:
 		case PRIM_INFO_DEACTIVATE:
@@ -449,25 +420,12 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 				/* attempt to allocate an Error Concealment Unit instance, if available */
 				lchan->ecu_state = osmo_ecu_init(trx, lchan2ecu_codec(lchan));
 
-				/* trx_chan_desc[] in scheduler.c uses the RSL_CHAN_OSMO_PDCH cbits
-				 * (0xc0) to indicate the need for PDTCH and PTCCH SAPI activation.
-				 * However, 0xc0 is a cbits pattern exclusively used for Osmocom style
-				 * dyn TS (a non-standard RSL Chan Activ mod); hence, for IPA style dyn
-				 * TS, the chan_nr will never reflect 0xc0 and we would omit the
-				 * PDTCH,PTTCH SAPIs. To properly de-/activate the PDTCH SAPIs in
-				 * scheduler.c, make sure the 0xc0 cbits are set for de-/activating PDTCH
-				 * lchans, i.e. both Osmocom and IPA style dyn TS. (For Osmocom style dyn
-				 * TS, the chan_nr typically already reflects 0xc0, while it doesn't for
-				 * IPA style.) */
-				if (lchan->type == GSM_LCHAN_PDTCH)
-					chan_nr = RSL_CHAN_OSMO_PDCH | (chan_nr & ~RSL_CHAN_NR_MASK);
-
 				/* activate dedicated channel */
-				trx_sched_set_lchan(&l1h->l1s, chan_nr, LID_DEDIC, true);
+				trx_sched_set_lchan(lchan, chan_nr, LID_DEDIC, true);
 				/* activate associated channel */
-				trx_sched_set_lchan(&l1h->l1s, chan_nr, LID_SACCH, true);
+				trx_sched_set_lchan(lchan, chan_nr, LID_SACCH, true);
 				/* set mode */
-				trx_sched_set_mode(&l1h->l1s, chan_nr,
+				trx_sched_set_mode(lchan->ts, chan_nr,
 					lchan->rsl_cmode, lchan->tch_mode,
 					lchan->tch.amr_mr.num_modes,
 					lchan->tch.amr_mr.bts_mode[0].mode,
@@ -476,21 +434,18 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 					lchan->tch.amr_mr.bts_mode[3].mode,
 					amr_get_initial_mode(lchan),
 					(lchan->ho.active == 1));
-				/* init lapdm */
-				lchan_init_lapdm(lchan);
 				/* set lchan active */
 				lchan_set_state(lchan, LCHAN_S_ACTIVE);
 				/* set initial ciphering */
-				l1if_set_ciphering(l1h, lchan, chan_nr, 0);
-				l1if_set_ciphering(l1h, lchan, chan_nr, 1);
+				l1if_set_ciphering(lchan, chan_nr, 0);
+				l1if_set_ciphering(lchan, chan_nr, 1);
 				if (lchan->encr.alg_id)
 					lchan->ciph_state = LCHAN_CIPH_RXTX_CONF;
 				else
 					lchan->ciph_state = LCHAN_CIPH_NONE;
 
 				/* confirm */
-				mph_info_chan_confirm(l1h, chan_nr,
-					PRIM_INFO_ACTIVATE, 0);
+				mph_info_chan_confirm(trx, chan_nr, PRIM_INFO_ACTIVATE, 0);
 				break;
 			}
 			if (l1sap->u.info.type == PRIM_INFO_MODIFY) {
@@ -499,7 +454,7 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 					osmo_ecu_destroy(lchan->ecu_state);
 				lchan->ecu_state = osmo_ecu_init(trx, lchan2ecu_codec(lchan));
 				/* change mode */
-				trx_sched_set_mode(&l1h->l1s, chan_nr,
+				trx_sched_set_mode(lchan->ts, chan_nr,
 					lchan->rsl_cmode, lchan->tch_mode,
 					lchan->tch.amr_mr.num_modes,
 					lchan->tch.amr_mr.bts_mode[0].mode,
@@ -527,8 +482,7 @@ int bts_model_l1sap_down(struct gsm_bts_trx *trx, struct osmo_phsap_prim *l1sap)
 				/* deactivate dedicated channel */
 				lchan_deactivate(lchan);
 				/* confirm only on dedicated channel */
-				mph_info_chan_confirm(l1h, chan_nr,
-					PRIM_INFO_DEACTIVATE, 0);
+				mph_info_chan_confirm(trx, chan_nr, PRIM_INFO_DEACTIVATE, 0);
 			}
 			break;
 		default:
@@ -570,21 +524,38 @@ int bts_model_apply_oml(struct gsm_bts *bts, struct msgb *msg,
 			struct tlv_parsed *new_attr, int kind, void *obj)
 {
 	struct abis_om_fom_hdr *foh = msgb_l3(msg);
-	int cause = 0;
+	struct gsm_abis_mo *mo = gsm_objclass2mo(bts, foh->obj_class, &foh->obj_inst);
+	struct nm_fsm_ev_setattr_data ev_data = {
+		.msg = msg,
+		.cause = 0,
+	};
+	int rc;
+
+	/* TODO: NM Object without FSM: */
+	switch (foh->obj_class) {
+	case NM_OC_GPRS_NSE:
+	case NM_OC_GPRS_CELL:
+	case NM_OC_GPRS_NSVC:
+		return oml_fom_ack_nack(ev_data.msg, ev_data.cause);
+	}
 
 	switch (foh->msg_type) {
 	case NM_MT_SET_BTS_ATTR:
-		cause = trx_set_bts(obj, new_attr);
+		ev_data.cause = trx_set_bts(obj, new_attr);
 		break;
 	case NM_MT_SET_RADIO_ATTR:
-		cause = trx_set_trx(obj);
+		ev_data.cause = trx_set_trx(obj);
 		break;
 	case NM_MT_SET_CHAN_ATTR:
-		cause = trx_set_ts(obj);
+		ev_data.cause = trx_set_ts(obj);
 		break;
 	}
 
-	return oml_fom_ack_nack(msg, cause);
+	rc = osmo_fsm_inst_dispatch(mo->fi,
+				    ev_data.cause == 0 ? NM_EV_SETATTR_ACK : NM_EV_SETATTR_NACK,
+				    &ev_data);
+	/* msgb ownsership is transferred to FSM if it received ev: */
+	return rc == 0 ? 1 : 0;
 }
 
 /* callback from OML */
@@ -619,7 +590,7 @@ int bts_model_opstart(struct gsm_bts *bts, struct gsm_abis_mo *mo,
 	case NM_OC_GPRS_NSE:
 	case NM_OC_GPRS_CELL:
 	case NM_OC_GPRS_NSVC:
-		oml_mo_state_chg(mo, NM_OPSTATE_ENABLED, NM_AVSTATE_OK);
+		oml_mo_state_chg(mo, NM_OPSTATE_ENABLED, NM_AVSTATE_OK, -1);
 		rc = oml_mo_opstart_ack(mo);
 		break;
 	default:
